from __future__ import annotations

from collections.abc import Iterator
from typing import Any

from pydantic import ValidationError

from pydantic_marshals.base.fields.base import MarshalField
from pydantic_marshals.base.models import FieldConverter
from pydantic_marshals.base.type_aliases import FieldType, TypeHint
from pydantic_marshals.contains.fields.constants import (
    ArbitraryConstantField,
    LiteralConstantField,
)
from pydantic_marshals.contains.fields.lists import strict_list_field_factory
from pydantic_marshals.contains.fields.nested import nested_field_factory
from pydantic_marshals.contains.fields.typed import GeneratedTypeField, TypedField
from pydantic_marshals.contains.fields.wildcards import (
    AnythingField,
    NothingField,
    SomethingField,
)
from pydantic_marshals.contains.type_aliases import TypeChecker


class AssertContainsModel(FieldConverter):
    field_types = (
        NothingField,
        SomethingField,
        AnythingField,
        LiteralConstantField,
        TypedField,
        GeneratedTypeField,
    )

    default_field_type = ArbitraryConstantField

    @classmethod
    def dynamic_field_types(cls) -> Iterator[type[MarshalField]]:
        yield nested_field_factory(cls.convert_to_field)
        yield strict_list_field_factory(cls.convert_to_type)

    @classmethod
    def convert_to_type(cls, source: TypeChecker) -> TypeHint:
        return cls.convert_field(source).generate_type()

    @classmethod
    def convert_to_field(cls, source: TypeChecker) -> FieldType:
        return cls.convert_field(source).generate_field()

    @classmethod
    def contains(cls, real: Any, expected: TypeChecker) -> None:
        cls.convert_field(expected).generate_root_model().model_validate(real)


def assert_contains(real: Any, expected: TypeChecker) -> None:
    try:
        AssertContainsModel.contains(real, expected)
    except ValidationError as e:
        raise AssertionError(str(e)) from None



from collections.abc import Callable
from typing import Final

from mypy.build import PRI_ALL
from mypy.mro import calculate_mro
from mypy.nodes import (
    GDEF,
    Block,
    ClassDef,
    MypyFile,
    SymbolTable,
    SymbolTableNode,
    TypeInfo,
)
from mypy.plugin import DynamicClassDefContext, Plugin

# TODO redo and generalize
base_model_qualname: Final = "pydantic_marshals.sqlalchemy.models.MappedModel"
type_matrix = {base_model_qualname}
methods = {"create", "extend", "as_patch"}

pydantic_base_model_qualname: Final = "pydantic_marshals.base.models.MarshalBaseModel"
stub_module_name: Final = "pydantic_marshals.mypy.magic"
stub_class_name: Final = f"{stub_module_name}.MappedModelStub"


class MarshalsPlugin(Plugin):
    def get_dynamic_class_hook(
        self,
        fullname: str,
    ) -> Callable[[DynamicClassDefContext], None] | None:
        class_fullname, method_name = fullname.rpartition(".")[::2]
        if (
            fullname == base_model_qualname
            or class_fullname in type_matrix
            and method_name in methods
        ):
            return add_info_hook
        return None

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        # Without PRI_ALL causes pydantic.mypy to not activate for some reason
        return [(PRI_ALL, stub_module_name, -1)]


def add_info_hook(ctx: DynamicClassDefContext) -> None:
    class_def = ClassDef(ctx.name, Block([]))
    class_def.fullname = ctx.api.qualified_name(ctx.name)
    info = TypeInfo(SymbolTable(), class_def, ctx.api.cur_mod_id)
    class_def.info = info
    bm = ctx.api.named_type(pydantic_base_model_qualname)
    pm = ctx.api.named_type(stub_class_name)
    info.bases = [bm, pm]
    info.fallback_to_any = True
    calculate_mro(info)

    ctx.api.add_symbol_table_node(ctx.name, SymbolTableNode(GDEF, info))
    type_matrix.add(class_def.fullname)


def plugin(_: str) -> type[Plugin]:
    return MarshalsPlugin



from __future__ import annotations

from typing import Annotated, Any, Generic, Literal, TypeVar

from pydantic import AfterValidator
from typing_extensions import Self

from pydantic_marshals.base.fields.base import MarshalField
from pydantic_marshals.base.type_aliases import TypeHint
from pydantic_marshals.contains.type_aliases import LiteralType


class LiteralConstantField(MarshalField):
    def __init__(self, constant: LiteralType) -> None:
        super().__init__()
        self.constant = constant

    @classmethod
    def convert(cls, constant: Any = None, *_: Any) -> Self | None:
        if isinstance(constant, LiteralType):  # type: ignore[misc, arg-type]
            # https://github.com/python/mypy/issues/12155 ^^^^^^^^^^^^^^^^^^^
            return cls(constant)
        return None

    def generate_type(self) -> TypeHint:
        # noinspection PyTypeHints
        return Literal[self.constant]  # pycharm bug


FieldType = TypeVar("FieldType")


class ArbitraryConstantField(MarshalField, Generic[FieldType]):
    def __init__(self, expected: FieldType) -> None:
        super().__init__()
        self.expected = expected
        self.type_ = type(expected)

    @classmethod
    def convert(cls, expected: Any = None, *_: Any) -> Self | None:
        if expected == expected:  # noqa: WPS312
            # check if equality operator is supported
            return cls(expected)
        return None

    def equality_validator(self, real: FieldType) -> FieldType:
        if real != self.expected:
            raise ValueError(f"should be equal to '{self.expected}'")
        return real

    def generate_type(self) -> TypeHint:
        return Annotated[self.type_, AfterValidator(self.equality_validator)]



from __future__ import annotations

from types import UnionType
from typing import Annotated, Any, Union, get_origin

from typing_extensions import Self

from pydantic_marshals.base.fields.base import MarshalField
from pydantic_marshals.base.type_aliases import TypeHint
from pydantic_marshals.contains.type_generators.base import BaseTypeGenerator


class TypedField(MarshalField):
    def __init__(self, type_: TypeHint) -> None:  # TODO optional
        super().__init__()
        self.type_ = type_

    @classmethod
    def convert(cls, source: Any = None, *_: Any) -> Self | None:
        if (
            source is not Any
            and isinstance(source, type)
            or get_origin(source) in {Annotated, Union, UnionType}
        ):
            return cls(source)
        return None

    def generate_type(self) -> TypeHint:
        return self.type_


class GeneratedTypeField(TypedField):
    @classmethod
    def convert(cls, source: Any = None, *_: Any) -> Self | None:
        if isinstance(source, BaseTypeGenerator):
            return cls(source.to_typehint())
        return None



from pydantic_marshals.base.composite import CompositeMarshalModel
from pydantic_marshals.base.fields.base import PatchDefault, PatchDefaultType
from pydantic_marshals.base.models import MarshalModel

__all__ = ("MarshalModel", "PatchDefault", "PatchDefaultType", "CompositeMarshalModel")



from __future__ import annotations

from typing import Any

from pydantic import BaseModel
from sqlalchemy.orm import Mapped, Relationship
from typing_extensions import Self

from pydantic_marshals.base.fields.base import PatchMarshalField
from pydantic_marshals.base.type_aliases import TypeHint
from pydantic_marshals.utils import ModeledType, is_subtype


class RelationshipField(PatchMarshalField):
    """
    Implementation of :py:class:`MarshalField` to use with SQLAlchemy's relationships
    A conversion model has to be specified, preferably one made with
    :py:class:`pydantic_marshals.models.sqlalchemy.MappedModel`
    """

    def __init__(
        self,
        mapped_relationship: Relationship[Any],
        model: type[BaseModel],
        nullable: bool = False,
        alias: str | None = None,
        patch: bool = False,
    ) -> None:
        super().__init__(alias, patch)
        self.relationship = mapped_relationship
        self.model = model
        self.nullable = nullable

    def as_patch(self) -> RelationshipField:
        return RelationshipField(
            mapped_relationship=self.relationship,
            model=self.model,
            nullable=self.nullable,
            alias=self.alias,
            patch=True,
        )

    @classmethod
    def convert(
        cls,
        mapped: Any = None,
        model: Any = None,
        nullable: Any = False,
        *_: Any,
    ) -> Self | None:
        if (
            isinstance(mapped, Relationship)
            and is_subtype(model, BaseModel)
            and isinstance(nullable, bool)
        ):
            return cls(mapped, model, nullable)
        return None

    def generate_name(self) -> str:
        return self.relationship.key

    def generate_type(self) -> TypeHint:
        collection_class = self.relationship.collection_class
        if collection_class is None:
            if self.nullable:
                return self.model | None
            return self.model
        if not isinstance(collection_class, type):
            raise RuntimeError(f"Collection is not a type: {collection_class}")
        if issubclass(collection_class, list):
            return list[self.model]  # type: ignore[name-defined]
        raise RuntimeError(f"Bad collection class: {collection_class}")


RelationshipType = (
    ModeledType[Relationship[Any]] | ModeledType[Mapped[Any]] | RelationshipField
)



from __future__ import annotations

from collections.abc import Callable
from typing import Any

from typing_extensions import Self

from pydantic_marshals.base.fields.base import MarshalField
from pydantic_marshals.base.type_aliases import TypeHint
from pydantic_marshals.contains.fields.typed import TypedField
from pydantic_marshals.contains.type_aliases import TypeChecker


def strict_list_field_factory(
    convert_type: Callable[[TypeChecker], TypeHint],
) -> type[MarshalField]:
    class StrictListFieldInner(TypedField):
        @classmethod
        def convert(cls, source: Any = None, *_: Any) -> Self | None:
            if isinstance(source, list):
                return cls(
                    tuple[  # type: ignore[misc]
                        tuple(convert_type(value) for value in source)
                    ],
                )
            return None

    return StrictListFieldInner



from __future__ import annotations

from collections.abc import Iterator
from typing import Any

from typing_extensions import Self

from pydantic_marshals.base.fields.base import MarshalField
from pydantic_marshals.base.type_aliases import TypeHint


class NothingField(MarshalField):
    @classmethod
    def convert(cls, source: Any = None, *_: Any) -> Self | None:
        if source is None:
            return cls()
        return None

    def generate_type(self) -> TypeHint:
        return type(None)

    def generate_field_data(self) -> Iterator[tuple[str, Any]]:
        yield "default", None


class SomethingField(MarshalField):
    @classmethod
    def convert(cls, source: Any = None, *_: Any) -> Self | None:
        if source is ...:
            return cls()
        return None

    def generate_type(self) -> TypeHint:
        return Any


class AnythingField(MarshalField):
    @classmethod
    def convert(cls, source: Any = None, *_: Any) -> Self | None:
        if source is Any:
            return cls()
        return None

    def generate_type(self) -> TypeHint:
        return Any

    def generate_field_data(self) -> Iterator[tuple[str, Any]]:
        yield "default", None



from __future__ import annotations

from collections.abc import Callable
from contextlib import suppress
from typing import Any

from pydantic import ConfigDict, create_model
from typing_extensions import Self

from pydantic_marshals.base.fields.base import MarshalField
from pydantic_marshals.base.type_aliases import FieldType
from pydantic_marshals.contains.fields.typed import TypedField
from pydantic_marshals.contains.type_aliases import TypeChecker


def nested_field_factory(  # noqa: N802
    convert_field: Callable[[TypeChecker], FieldType],
) -> type[MarshalField]:
    class NestedFieldInner(TypedField):
        @classmethod
        def convert(cls, source: Any = None, *_: Any) -> Self | None:
            if isinstance(source, dict):
                with suppress(RuntimeError):
                    fields: dict[str, FieldType] = {
                        key: convert_field(value) for key, value in source.items()
                    }  # TODO check if `isinstance(value, TypeChecker)`
                    return cls(
                        create_model(  # type: ignore[call-overload]
                            "Model",
                            **fields,
                            __config__=ConfigDict(  # TODO maybe move to `contains`
                                from_attributes=True,
                                arbitrary_types_allowed=True,
                            ),
                        ),
                    )
            return None

    return NestedFieldInner



from __future__ import annotations

from enum import Enum
from types import EllipsisType
from typing import TypeAlias

from pydantic_marshals.contains.type_generators.base import BaseTypeGenerator

LiteralType: TypeAlias = bool | int | float | bytes | str | Enum
TypeChecker: TypeAlias = (
    None
    | type
    | dict
    | list
    | type[LiteralType]
    | LiteralType
    | EllipsisType
    | BaseTypeGenerator
)



from typing import Annotated, Any

from pydantic import AfterValidator

from pydantic_marshals.base.type_aliases import TypeHint


class BaseTypeGenerator:
    """
    Helper interface to make creating custom types for pydantic easier
    Subclasses must override :py:attr:`data_type` and :py:meth:`_validate`

    See also: https://docs.pydantic.dev/2.0/usage/types/custom
    """

    def __init__(self, data_type: TypeHint) -> None:
        self.data_type = data_type

    def _validate(self, data: Any) -> None:
        """
        Method to validate `data`. Should throw errors on fails, same as validators in
        pydantic: https://docs.pydantic.dev/2.0/usage/validators/
        """
        raise NotImplementedError

    def validate(self, data: Any) -> Any | None:
        self._validate(data)
        return data

    def to_typehint(self) -> TypeHint:
        """
        Converts type generator into a type hint, usable as a custom type with pydantic
        https://docs.pydantic.dev/2.0/usage/types/custom
        """
        return Annotated[self.data_type, AfterValidator(self.validate)]



from types import NoneType, UnionType
from typing import Any, TypeVar, get_args, get_origin

from pydantic import BaseModel

T = TypeVar("T")
ModeledType = tuple[T, type[BaseModel]]


def is_subtype(maybe_type: Any, klass: type) -> bool:
    return isinstance(maybe_type, type) and issubclass(maybe_type, klass)


def is_optional(annotation: Any) -> bool:
    return get_origin(annotation) == UnionType and NoneType in get_args(annotation)



from __future__ import annotations

from collections.abc import Iterator
from enum import Enum
from typing import Any
from warnings import warn

from pydantic_core import PydanticUndefined, PydanticUndefinedType
from sqlalchemy.orm import Mapped, MappedColumn
from sqlalchemy.sql.schema import Column, ScalarElementColumnDefault
from sqlalchemy.sql.sqltypes import String
from typing_extensions import Self

from pydantic_marshals.base.fields.base import PatchMarshalField
from pydantic_marshals.base.type_aliases import TypeHint
from pydantic_marshals.utils import is_subtype


class ColumnField(PatchMarshalField):
    """
    Implementation of :py:class:`MarshalField` to use with SQLAlchemy's columns
    This will only work with 2.0-style MappedColumns
    """

    def __init__(
        self,
        mapped_column: MappedColumn[Any],
        type_: TypeHint | None = None,
        alias: str | None = None,
        patch: bool = False,
    ) -> None:
        super().__init__(alias, patch)
        self.mapped = mapped_column
        self.type_override = type_

    def as_patch(self) -> ColumnField:
        return ColumnField(
            mapped_column=self.mapped,
            type_=self.type_override,
            alias=self.alias,
            patch=True,
        )

    @classmethod
    def convert(
        cls,
        source: Any = None,
        type_: Any = None,
        alias: Any = None,
        *_: Any,
    ) -> Self | None:
        if isinstance(source, MappedColumn) and (
            type_ is None or isinstance(type_, TypeHint)
        ):
            if isinstance(alias, str):
                return cls(
                    mapped_column=source,
                    type_=type_,
                    alias=alias,
                )
            return cls(mapped_column=source, type_=type_)
        return None

    @property
    def column(self) -> Column[Any]:
        return self.mapped.column

    def generate_name(self) -> str:
        return self.column.name

    def generate_type(self) -> TypeHint:
        if self.type_override is not None:
            return self.type_override
        type_: TypeHint = self.column.type.python_type
        if self.column.nullable:
            return type_ | None
        return type_

    def generate_default(self) -> Any | None | PydanticUndefinedType:
        if self.column.default is None:
            return None if self.column.nullable else PydanticUndefined
        if isinstance(self.column.default, ScalarElementColumnDefault):
            return self.column.default.arg
        warn(f"Default: {self.column.default} is not supported, skipping")
        return PydanticUndefined

    def generate_field_data(self) -> Iterator[tuple[str, Any]]:
        yield from super().generate_field_data()

        column_type = self.column.type
        if isinstance(column_type, String) and not is_subtype(
            column_type.python_type,
            Enum,
        ):  # enums are kept by name in the database
            yield "max_length", column_type.length


ColumnType = (
    MappedColumn[Any]
    | tuple[MappedColumn[Any], TypeHint]
    | tuple[MappedColumn[Any], TypeHint, str]
    | Mapped[Any]
    | tuple[Mapped[Any], TypeHint]
    | tuple[Mapped[Any], TypeHint, str]
    | ColumnField
)



from __future__ import annotations

from types import GenericAlias, UnionType
from typing import Any, Literal, TypeAlias

from pydantic.fields import FieldInfo

BaseGenericAlias: TypeAlias = type(  # type: ignore[name-defined]
    Literal[0],
).__base__.__base__
SpecialForm: TypeAlias = type(Any)  # type: ignore[valid-type]
TypeHint: TypeAlias = (
    type  # type: ignore[valid-type]
    | GenericAlias
    | UnionType
    | BaseGenericAlias
    | SpecialForm
    | type(None)
)

FieldType: TypeAlias = TypeHint | tuple[TypeHint, Any | FieldInfo]



from __future__ import annotations

from typing import Any, get_origin

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from pydantic_marshals.base.models import MarshalBaseModel
from pydantic_marshals.utils import is_optional, is_subtype


class CompositeMarshalModel(MarshalBaseModel):
    @classmethod
    def generate_marshal_model_name(cls) -> str:
        return f"{cls.__name__}Marshal"

    @classmethod
    def convert_field(cls, field: FieldInfo) -> tuple[Any, Any]:
        if len(field.metadata) == 1 and is_subtype(field.metadata[0], MarshalBaseModel):
            if is_subtype(get_origin(field.annotation), list):
                return list[field.metadata[0]], field.default  # type: ignore
            if is_optional(field.annotation):
                return field.metadata[0] | None, field.default
            return field.metadata[0], field.default
        if field.annotation is None:
            raise TypeError("Annotation is somehow None")
        return field.annotation, field

    @classmethod
    def build_marshal(cls) -> type[BaseModel]:
        return create_model(  # type: ignore[call-overload, no-any-return]
            cls.generate_marshal_model_name(),
            **{
                file_name: cls.convert_field(field)
                for file_name, field in cls.model_fields.items()
            },
        )



from __future__ import annotations

from collections.abc import Callable
from typing import Any, get_type_hints

from typing_extensions import Self

from pydantic_marshals.base.fields.base import PatchMarshalField
from pydantic_marshals.base.type_aliases import TypeHint
from pydantic_marshals.utils import ModeledType


class PropertyField(PatchMarshalField):
    """
    Implementation of :py:class:`MarshalField` to use with properties
    Can be used directly or with an added type hit override
    """

    def __init__(
        self,
        mapped_property: property,
        type_: TypeHint | None = None,
        alias: str | None = None,
        patch: bool = False,
    ) -> None:
        super().__init__(alias, patch)
        self.mapped_property = mapped_property

        if self.mapped_property.fget is None:
            raise RuntimeError("Property's fget is None somehow")
        self.getter: Callable[[Any], Any] = self.mapped_property.fget

        self.name: str = self.getter.__name__
        self.type_: TypeHint = type_ or get_type_hints(self.getter).get("return", Any)

    def as_patch(self) -> PropertyField:
        return PropertyField(self.mapped_property, self.type_, self.alias, patch=True)

    @classmethod
    def convert(cls, mapped: Any = None, type_: Any = None, *_: Any) -> Self | None:
        if isinstance(mapped, property) and isinstance(type_, TypeHint):
            return cls(mapped, type_)
        return None

    def generate_name(self) -> str:
        return self.name

    def generate_type(self) -> TypeHint:
        return self.type_


PropertyType = (
    property
    | ModeledType[property]
    | Callable[[Any], Any]
    | ModeledType[Callable[[Any], Any]]
    | PropertyField
)



from pydantic_marshals.contains.models import AssertContainsModel, assert_contains
from pydantic_marshals.contains.type_aliases import TypeChecker
from pydantic_marshals.contains.type_generators.collections import (
    UnorderedLiteralCollection,
)

__all__ = (
    "AssertContainsModel",
    "assert_contains",
    "TypeChecker",
    "UnorderedLiteralCollection",
)



from collections.abc import Iterable

from pydantic_marshals.base.type_aliases import TypeHint
from pydantic_marshals.contains.type_aliases import LiteralType
from pydantic_marshals.contains.type_generators.base import BaseTypeGenerator


class UnorderedLiteralCollection(BaseTypeGenerator):
    def __init__(
        self,
        items: Iterable[LiteralType],
        data_type: TypeHint = LiteralType,
        check_extra: bool = True,
        check_repeats: bool = True,
    ) -> None:
        """
        :param items: set of literal values to search for
        :param check_extra: (default True) determines if extra values are allowed
        :param check_repeats: (default True) determines if repeating values are allowed
        """
        super().__init__(Iterable[data_type])
        self.items: set[LiteralType] = set(items)
        self.check_extra = check_extra
        self.check_repeats = check_repeats

    def _validate(self, data: Iterable[LiteralType]) -> None:
        data_set: set[LiteralType] = set(data)

        missing: set[LiteralType] = self.items - data_set
        if missing:
            raise ValueError(f"items missing: {missing}")

        if self.check_extra:
            extra: set[LiteralType] = data_set - self.items
            if extra:
                raise ValueError(f"extra items found: {extra}")

        if self.check_repeats:
            repeats: set[LiteralType] = set()
            seen: set[LiteralType] = set()

            for item in data:
                if item in seen:
                    repeats.add(item)
                else:
                    seen.add(item)

            if repeats:
                raise ValueError(f"repeating items found: {repeats}")



from __future__ import annotations

from collections.abc import Iterator, Sequence

from pydantic import BaseModel
from typing_extensions import Self

from pydantic_marshals.base.fields.base import MarshalField, PatchMarshalField
from pydantic_marshals.base.fields.properties import PropertyField, PropertyType
from pydantic_marshals.base.models import MarshalModel
from pydantic_marshals.base.type_aliases import FieldType
from pydantic_marshals.sqlalchemy.fields.columns import ColumnField, ColumnType
from pydantic_marshals.sqlalchemy.fields.relationships import (
    RelationshipField,
    RelationshipType,
)


class MappedModel(MarshalModel):
    """
    Implementation of :py:class:`MarshalModel` to use inside SQLAlchemy's
    (ORM style) table definitions (i.e. classes inherited from `Base`).

    Right now it can accept and convert into fields:
    - columns (constructed with ``mapped_column()``)
    - relationships (constructed with ``relationship()``)
    - properties (constructed with built-in ``@property()``)
    - properties with overriden type (see :py:class:`PropertyField`)

    An alternative constructor :py:meth:`create` is available to make this happen
    """

    field_types = (
        ColumnField,
        RelationshipField,
        PropertyField,
    )

    @classmethod
    def create(
        cls,
        columns: Sequence[ColumnType] = (),
        relationships: Sequence[RelationshipType] = (),
        properties: Sequence[PropertyType] = (),
        bases: Sequence[type[BaseModel]] = (),
        includes: Sequence[MarshalModel] = (),
        extra_fields: dict[str, FieldType] | None = None,
    ) -> Self:
        return cls(
            *cls.convert_fields(*columns),
            *cls.convert_fields(*relationships),
            *cls.convert_fields(*properties),
            *(field for model in includes for field in model.fields),
            bases=list(bases) + [base for model in includes for base in model.bases],
            extra_fields={} if extra_fields is None else extra_fields,
        )

    def extend(
        self,
        columns: Sequence[ColumnType] = (),
        relationships: Sequence[RelationshipType] = (),
        properties: Sequence[PropertyType] = (),
        bases: Sequence[type[BaseModel]] = (),
        includes: Sequence[MarshalModel] = (),
        extra_fields: dict[str, FieldType] | None = None,
    ) -> Self:
        return self.create(
            columns=columns,
            relationships=relationships,
            properties=properties,
            bases=bases,
            includes=(self, *includes),
            extra_fields={
                **self.extra_fields,
                **({} if extra_fields is None else extra_fields)
            },
        )

    def patch_fields(self) -> Iterator[MarshalField]:
        for field in self.fields:
            if isinstance(field, PatchMarshalField):
                yield field.as_patch()
            else:
                yield field

    def as_patch(self) -> Self:
        return type(self)(*self.patch_fields(), bases=self.bases)



from __future__ import annotations

from collections.abc import Iterator
from enum import Enum
from typing import Any, ClassVar, Generic, TypeVar

from pydantic import ConfigDict, RootModel
from pydantic.fields import Field
from pydantic_core import PydanticUndefined, PydanticUndefinedType
from typing_extensions import Self

from pydantic_marshals.base.type_aliases import FieldType, TypeHint

T = TypeVar("T")


class MarshalRootModel(RootModel[T], Generic[T]):
    model_config = ConfigDict(
        from_attributes=True,
        populate_by_name=True,
        arbitrary_types_allowed=True,
    )


class MarshalField:
    """
    Basic boilerplate class for all pydantic-marshals' models' fields.
    This is an interface, and it requires implementing all methods to function
    and to be used in :py:class:`pydantic_marshals.models.base.MarshalModel`
    """

    marshal_root_model: ClassVar[type[RootModel[Any]]] = MarshalRootModel
    """Base root model class. Subclasses of :py:class:`MarshalRootModel` are recommended"""

    def __init__(self, alias: str | None = None) -> None:
        """
        :param alias: same as Field(alias=...), can be None for no alias
        """
        self.alias = alias

    @classmethod
    def convert(cls, *source: Any) -> Self | None:
        """
        Convert something into a field.
        If conversion is not possible, this method should return None
        """
        raise NotImplementedError

    def generate_name(self) -> str:
        """
        Generates the name for the field, used in
        :py:meth:`pydantic_marshals.models.base.MarshalModel.generate_model`
        """
        raise NotImplementedError

    def generate_type(self) -> TypeHint:
        """
        Generates the type annotation for the field, used in
        :py:meth:`pydantic_marshals.models.base.MarshalModel.generate_model`
        """
        raise NotImplementedError

    # TODO type with TypedDict? Use _FieldInfoInputs?
    def generate_field_data(self) -> Iterator[tuple[str, Any]]:
        """
        Generates field data (kwargs for :py:func:`pydantic.fields.Field`),
        and returns it in the for of an Iterator,
        compatible with the ``dict[str, Any]`` constructor

        Kwarg names and types can be found in
        :py:class:`pydantic.fields._FieldInfoInputs`
        """
        if self.alias is not None:
            yield "alias", self.alias

    def generate_field(self) -> FieldType:
        """
        Generates field info for the field, used in
        :py:meth:`pydantic_marshals.models.base.MarshalModel.generate_model`
        """
        return (
            self.generate_type(),
            Field(**dict(self.generate_field_data())),
        )

    def generate_root_model(self) -> type[RootModel[Any]]:
        # TODO maybe move to `contains`
        return self.marshal_root_model[self.generate_type()]  # type: ignore[no-any-return, index]


class PatchDefaultType(Enum):
    PatchDefault = object()


PatchDefault = PatchDefaultType.PatchDefault


class PatchMarshalField(MarshalField):
    def __init__(self, alias: str | None = None, patch: bool = False) -> None:
        """
        :param alias: same as Field(alias=...), can be None for no alias
        :param patch: use PatchDefault as a default for the filed
        """
        super().__init__(alias=alias)
        self.patch = patch

    def as_patch(self) -> Self:
        """
        Creates the same field (copy), but with patch=True
        """
        raise NotImplementedError

    def generate_default(self) -> Any | None | PydanticUndefinedType:
        """
        Generates the default for the field, used in
        :py:meth:`pydantic_marshals.models.base.MarshalModel.generate_field_data`
        """
        return PydanticUndefined

    def generate_field_data(self) -> Iterator[tuple[str, Any]]:
        yield "default", PatchDefault if self.patch else self.generate_default()

        yield from super().generate_field_data()
